;
; asm.asm -- file containing all the manually written assembler code
;
; No, not modular at all, but it makes re-building easier when you
; don't have a makefile.
;
; Includes routines to:
;	copy TRS80 screen to CGA screen
;	install VGA font
;
; On the CGA screen copy:
;
; A 386 version would be easier to write, faster and neater.
; However, I will restrain myself to 8086.
; The components we have are the TRS80 display memory, the shadow
; screen, the character ROM data and the PC display.  We can
; guarantee that the character ROM and shadow screen will be in
; the same 64K segment.  We need three segment registers and
; 4 index registers:
;
;	ROM	ds:si
;	shadow	ds:bp
;	trs80	es:bx
;	pc	es:di	steal es when necessary
;
; We'll use ch and cl for row and column counters.

ASM_TEXT	segment	byte public 'CODE'
DGROUP	group	_DATA,_BSS
	assume	cs:ASM_TEXT,ds:DGROUP
ASM_TEXT	ends

_DATA	segment word public 'DATA'
_msx	dw	0
_msy	dw	0
msbut	db	0
msaddr	dw	0

mheight	equ	16
mwidth	equ	9
mbywid	equ	(mwidth + 7) / 8
msave	db	(mheight + 1) * mbywid dup (0)

mhx	equ	2
mhy	equ	1

mshape	db	00h,00h
	db	40h,00h
	db	60h,00h
	db	70h,00h
	db	78h,00h
	db	7ch,00h
	db	7eh,00h
	db	7fh,00h
	db	7fh,80h
	db	7fh,00h
	db	7ch,00h
	db	46h,00h
	db	06h,00h
	db	03h,00h
	db	03h,00h
	db	00h,00h

mmask	db	03fh,0ffh
	db	01fh,0ffh
	db	00fh,0ffh
	db	007h,0ffh
	db	003h,0ffh
	db	001h,0ffh
	db	000h,0ffh
	db	000h,07fh
	db	000h,03fh
	db	000h,07fh
	db	001h,0ffh
	db	010h,0ffh
	db	0b0h,0ffh
	db	0f8h,07fh
	db	0f8h,07fh
	db	0fch,0ffh

_DATA	ends

_BSS	segment word public 'BSS'
_BSS	ends

ASM_TEXT	segment	byte public 'CODE'

	public	_cgascreen32
	public	_cgascreen64
	public	_instcharrom

	extrn	_trs_char:near
	extrn	_trs_char32:near
	extrn	_shadow:near
	extrn	_screen_mem:far
	extrn	_fullupdate:word

; for 32 characters per line mode
_cgascreen32 proc far
; save some registers (probably could get away with saving fewer)
	push	bp
	push	bx
	push	cx
	push	di
	push	es
	push	si
; get the vital information from global variables
	mov	si,offset DGROUP:_trs_char32
	mov     bp,offset DGROUP:_shadow
	les     bx,dword ptr ds:_screen_mem
	mov	di,8
	mov	dx,word ptr ds:_fullupdate
; now we're ready to start the main loop
	cld
	mov	ch,16
rowlp:	mov	cl,32
collp:	mov	al,byte ptr es:[bx]	; byte from display
	or	dx,dx
	jnz	nocmp
	mov	ah,byte ptr ds:[bp]	; shadow screen
	cmp	ah,al
	jz	noupdate		; nothing to do if they're the same
nocmp:	mov	byte ptr ds:[bp],al	; update the shadow
	push    si			; save ROM pointer
	push	es			; and segment of TRS80 memory
	xor	ah,ah
	shl	ax,1
	shl	ax,1
	shl	ax,1
	add	si,ax
	shl	ax,1
	add	si,ax			; SI points to ROM data for char
	mov	ax,0b800h
	mov	es,ax			; segment of CGA screen
	lodsw				; dump the 24 bytes (12 lines x 2)
	mov	word ptr es:[di],ax
	lodsw
	mov	word ptr es:[di+2000h],ax
	lodsw
	mov	word ptr es:[di+80],ax
	lodsw
	mov	word ptr es:[di+2000h+80],ax
	lodsw
	mov	word ptr es:[di+160],ax
	lodsw
	mov	word ptr es:[di+2000h+160],ax
	lodsw
	mov	word ptr es:[di+240],ax
	lodsw
	mov	word ptr es:[di+2000h+240],ax
	lodsw
	mov	word ptr es:[di+320],ax
	lodsw
	mov	word ptr es:[di+2000h+320],ax
	lodsw
	mov	word ptr es:[di+400],ax
	lodsw
	mov	word ptr es:[di+2000h+400],ax
	pop	es
	pop     si
noupdate:
	inc	di
	inc	di                      ; next PC screen location
	inc	bp
	inc	bp                      ; next shadow position
	inc	bx
	inc	bx			; next trs80 memory location
	dec     cl
	jnz     collp
	add	di,480-32*2
	dec	ch
	jz	done
	jmp	rowlp
done:	mov	word ptr ds:_fullupdate,0
; clean up and go back
	pop	si
	pop	es
	pop	di
	pop	cx
	pop	bx
	pop	bp
	ret
_cgascreen32	endp

; for 64 characters per line mode
_cgascreen64 proc far
; save some registers (probably could get away with saving fewer)
	push	bp
	push	bx
	push	cx
	push	di
	push	es
	push	si
; get the vital information from global variables
	mov	si,offset DGROUP:_trs_char
	mov     bp,offset DGROUP:_shadow
	les     bx,dword ptr ds:_screen_mem
	mov	di,8
	mov	dx,word ptr ds:_fullupdate
; now we're ready to start the main loop
	cld
	mov	ch,16
rowlp2:	mov	cl,64
collp2:	mov	al,byte ptr es:[bx]	; byte from display
	or	dx,dx
	jnz	nocmp2
	mov	ah,byte ptr ds:[bp]	; shadow screen
	cmp	ah,al
	jz	noupdate2		; nothing to do if they're the same
nocmp2:	mov	byte ptr ds:[bp],al	; update the shadow
	push    si			; save ROM pointer
	push	es			; and segment of TRS80 memory
	xor	ah,ah
	shl	ax,1
	shl	ax,1
	add	si,ax
	shl	ax,1
	add	si,ax			; SI points to ROM data for char
	mov	ax,0b800h
	mov	es,ax			; segment of CGA screen
	lodsw				; dump the 12 bytes
	mov	byte ptr es:[di],al
	mov	byte ptr es:[di+2000h],ah
	lodsw
	mov	byte ptr es:[di+80],al
	mov	byte ptr es:[di+2000h+80],ah
	lodsw
	mov	byte ptr es:[di+160],al
	mov	byte ptr es:[di+2000h+160],ah
	lodsw
	mov	byte ptr es:[di+240],al
	mov	byte ptr es:[di+2000h+240],ah
	lodsw
	mov	byte ptr es:[di+320],al
	mov	byte ptr es:[di+2000h+320],ah
	lodsw
	mov	byte ptr es:[di+400],al
	mov	byte ptr es:[di+2000h+400],ah
	pop	es
	pop     si
noupdate2:
	inc	di                      ; next PC screen location
	inc	bp                      ; next shadow position
	inc	bx			; next trs80 memory location
	dec     cl
	jnz     collp2
	add	di,480-64
	dec	ch
	jnz	rowlp2
	mov	word ptr ds:_fullupdate,0
; clean up and go back
	pop	si
	pop	es
	pop	di
	pop	cx
	pop	bx
	pop	bp
	ret
_cgascreen64	endp

; install the TRS80 character ROM as a VGA font.
; I would have done this in C, but it seems that Turbo C 2.0
; has a bug in its intr() routine in the large model.

_instcharrom	proc far
	push	bp
	mov	bp,offset DGROUP:_trs_char
	mov	ax,ds
	mov	es,ax
	mov	cx,256
	mov	dx,0
	mov	bx,0c00h
	mov	ax,1110h
	int	10h
	pop	bp
	ret
_instcharrom endp

ASM_TEXT ends
	end

