/*
 * loadbba.c -- load a BASIC program stored in "binary" format.
 *
 * The binary format is the same as a BASIC program is stored on tape
 * and in memory.  Consequently, we just have to put it in place and
 * adjust a few pointers.
 *
 * Major hack for now that has us load "basic.snp" and then the
 * BASIC program.  Will figure out how to do it elegantly later!
 */

#include <stdio.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "snap.h"
#include "load.h"

unsigned int loadptr;

static unsigned char getpoke(struct trs80*, FILE*);

#define echeck()	if (loaderr) return 0

int loadbba(struct trs80* t, char* fname)
{
	FILE*	fp;
	int		i;
	int		ch;
	unsigned int prevptr;

	if (!(fp = path_fopen(fname, "rb")))
		return 0;

	if (!restoresnap(t, "basic.snp"))
		return 0;

	/* get pointer to start of BASIC program */
	prevptr = loadptr = Peek16(0x40a4);

	if (fgetc(fp) != 255) {
		loaderr = 1;
		strcpy(loaderr_str, "Not a binary BASIC file.");
		fclose(fp);
		return 0;
	}

	for (;;) {
		/* skip over pointer to next line # */
		/* if == 0, then we're done with the loading */
		ch = getpoke(t, fp);
		echeck();
		i = getpoke(t, fp);
		echeck();
		if (ch == 0 && i == 0)
			break;

		getpoke(t, fp);	/* grab line # */
		echeck();
		getpoke(t, fp);	/* 2 bytes! */
		echeck();

		/* read line until null terminating byte */
		while (getpoke(t, fp) != 0)
			echeck();

		/* point to next line */
		Poke16(prevptr, loadptr);
		prevptr = loadptr;
	}

	/* now set up a few soft pointers */
	Poke16(0x40f9, loadptr);	/* end of BASIC program pointer */
	Poke16(0x40fb, loadptr);	/* start of array variables pointer */
	Poke16(0x40fd, loadptr);	/* start of free memory pointer */

	fclose(fp);
	return 1;
}

static unsigned char getpoke(struct trs80* t, FILE* fp)
{
	int	ch;

	if ((ch = fgetc(fp)) == EOF) {
		ueof();
		return 0;
	}
	Poke(loadptr++, ch & 255);
	return ch;
}
