/*
 * loadbin -- load  a model III binary file (.3bn) into memory.
 *
 * The format of the .3bn files is just that of tape data written to disk.
 */
#include <stdio.h>
#include <string.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "load.h"

int loadbin(struct trs80* t, char* fname)
{
	FILE*		in;
	unsigned char	ch;
	unsigned int	addr;
	int				len;
	int				i;

	if (!(in = path_fopen(fname, "rb")))
		return 0;

	if (fgetc(in) != 0x55) {
		fclose(in);
		loaderr = 1;
		strcpy(loaderr_str, "Missing block header byte of 0x55h\n");
		return 0;
	}

	for (i = 0; i < 6; i++) {	/* the file name (6 characters long) */
		fgetc(in);
	}

	for (;;) {
		ch = fgetc(in);	/* block header byte */
		if (ch == 0x3c) {	/* read a data block */
			len = fgetc(in);/* block length */
			if (len == 0)
				len = 256;
			addr = fgetc(in);
			addr += fgetc(in) << 8;	/* start address */
			for (i = 0; i < len; i++, addr++)
				Poke(addr, fgetc(in));
			fgetc(in);	/* checksum */
		}
		else if (ch == 0x78) {	/* final block */
			addr = fgetc(in);
			addr += fgetc(in) << 8; /* execution address */
			fclose(in);
			t->z80.reg.PC = addr;
			return 1;
		}
		else {
			loaderr = 1;
			strcpy(loaderr_str, "Bad block number.");
			return 0;
		}
	}
}
