/*
 * loadcmd.c -- load a ".cmd" file into the simulated TRS80.  Can't be
 * sure that this is correct due to a shortage of ".cmd" files, but it
 * _seems_ pretty good.
 */

#include <stdio.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "load.h"

static int skip(FILE*, int);
static int getlowhi(FILE*, int*);
static int gethilow(FILE*, int*);
static FILE* fp;

#define echeck()	if (loaderr) goto err

int loadcmd(struct trs80* t, char* fname)
{
	int	n;
	int len;
	int i;
	unsigned int addr;
	int ch;

	if (!(fp = path_fopen(fname, "rb")))
		return 0;

	ch = fgetc(fp);

	if (ch == 1)
		ungetc(ch, fp);
	else if (ch != 5) {
		fclose(fp);
		loaderr = 1;
		strcpy(loaderr_str, "No magic byte of 5, not a .cmd file.");
		return 0;
	}
	else if (!skip(fp, fgetc(fp))) {
		fclose(fp);
		return 0;
	}

	for (;;) {
		if (!gethilow(fp, &n)) { fclose(fp); return 0; }
		if (n == 514)
			break;
		if (n > 256)
			n -= 256;
		n -= 2;
		if (n == 0)
			n = 256;
		else if (n < 0)
			n += 256;
		if (!getlowhi(fp, &addr)) { fclose(fp); return 0; }
		for (i = 0; i < n; i++)
			Poke(addr++, fgetc(fp));
	}

	if (!getlowhi(fp, &addr)) { fclose(fp); return 0; }
	fclose(fp);
	t->z80.reg.PC = addr;
	return 1;
}

static int skip(FILE* fp, int n)
{
	while (n-- > 0)
		if (fgetc(fp) == EOF) {
			ueof();
			return 0;
		}

	return 1;
}

static int gethilow(FILE* fp, int* val)
{
	int hi;
	int low;

	hi = fgetc(fp);
	low = fgetc(fp);

	if (hi == EOF || low == EOF) {
		ueof();
		return 0;
	}

	*val =  hi * 256 + low;

	return 1;
}

static int getlowhi(FILE* fp, int* val)
{
	int hi;
	int low;

	low = fgetc(fp);
	hi = fgetc(fp);

	if (low == EOF || hi == EOF) {
		ueof();
		return 0;
	}

	*val = hi * 256 + low;

	return 1;
}
