/*
 * menu.c -- display a menu so the user may change configuration
 */

#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <graphics.h>

#include "mtype.h"
/* need these 2 'cause Turbo C can be lame */
#include "z80.h"
#include "trs80.h"

#include "trs80pc.h"

static char* charset_name[] = { "MODEL 1", "model 3", "model 1" };
static char* keyboard_name[] = { "WYSIWYG", "physical" };
static char* colour_name[] = { "blue", "green", "cyan", "red", "violet",
	"brown", "white", "grey", "bright blue", "bright green", "bright cyan",
	"bright red", "bright violet", "yellow", "bright white" };
static char* sound_name[] = { "off", "on" };
static char* speed_name[] = { "Model 1", "Model 3", "Flat out", "Half" };

/* Note: a re-ordering of the menu[] array will require a change to the
 * M_ #defines in trs80pc.h
 */
struct menu {
	int active;
	char* name;
	int	low, high;
	int* val;
	char** val_names;
	void (*change)();
} menu[] = {
	{ 1, "Display",  0,  2, &charset, charset_name,  change_charset },
	{ 1, "Keyboard", 0,  1, &physmap, keyboard_name, change_keyboard },
	{ 1, "Colour",   1, 15, &colour,  colour_name,   change_colour },
	{ 1, "Sound",    0,  1, &sounds,  sound_name,    0 },
	{ 1, "Timing",	 0,  3, &speed,   speed_name,    change_speed },
	{ 1, "Return to TRS-80",	 0,  0, 0, 0, 0 },
	{ 1, "Quit",     0,  0, 0, 0, 0 }
};
#define MENU_SIZE	(sizeof(menu) / sizeof(struct menu))

char merr[32];
static int msel = 0;

static void draw_menu(int, int, int);
static int menu_change(int, int);

#define K_UP	(1)
#define K_LEFT	(2)
#define K_RIGHT	(3)
#define K_DOWN	(4)

void menu_active(int n, int active)
{
	menu[n].active = active;
}

int do_menu(int dw, int trs_menu)
{
	char ch, msave = merr[0];
	int i;

	for (;;) {
		merr[0] = msave;
		draw_menu(msel, dw, trs_menu);
		merr[0] = '\0';
		if ((ch = getch()) == 0) {
			switch (getch()) {
			case 0x48: ch = K_UP; break;
			case 0x50: ch = K_DOWN; break;
			case 0x4b: ch = K_LEFT; break;
			case 0x4d: ch = K_RIGHT; break;
			default: break;
			}
		}
		switch (ch) {
		case K_DOWN:
			do {
				msel++;
				if (msel >= MENU_SIZE)
					msel = 0;
			} while (!menu[msel].active);
			break;
		case K_UP:
			do {
				msel--;
				if (msel < 0)
					msel = MENU_SIZE - 1;
			} while (!menu[msel].active);
			break;
		case ' ':
		case 13:
		case '\t':
		case K_RIGHT:
			switch (menu_change(msel, 1)) {
			case 1: return 1;
			case 2: return 0;
			default: break;
			}
			break;
		case K_LEFT:
			switch (menu_change(msel, -1)) {
			case 1: return 1;
			case 2: return 0;
			default: break;
			}
			break;
		case 27:
			return 0;
		default:
			/* handle accelerator keys */
			if (ch >= 'a' && ch <= 'z')
				ch -= 'a' - 'A';
			for (i = 0; i < MENU_SIZE; i++) {
				if (menu[i].name[0] == ch && menu[i].active)
					switch (menu_change(i, 1)) {
					case 1: return 1;
					case 2: return 0;
					default: break;
					}
			}
			break;
		}
	}
}

static void topline(int, int);
static void botline(int, int);

#define MENWID 28

static void draw_menu(int sel, int doublewide, int trs_menu)
{
	int i, x, m, y, sw;

	sw = doublewide ? 20 : 40;

	y = 1;

	if (merr[0]) { /* display error message */
		i = strlen(merr);
		x = sw - (i + 4) / 2;
		gotoxy(x, y);
		topline(i + 4, trs_menu);
		y++;
		gotoxy(x, y);
		printf("%c %s %c", trs_menu ? 149 : 186, merr, trs_menu ? 170 : 186);
		y++;
		gotoxy(x, y);
		botline(i + 4, trs_menu);
		y++;
	}

	x = sw - MENWID / 2;

	gotoxy(x, y);
	topline(MENWID, trs_menu);
	y++;

	for (i = m = 0; i < MENU_SIZE; i++) {
		if (!menu[i].active)
			continue;
		gotoxy(x, y);
		printf("%c%c ", trs_menu ? 149 : 186, i == sel ? '>' : ' ');
		printf("%-10s ", menu[i].name);
		if (menu[i].val) {
			if (menu[i].val_names)
				printf("%13s", menu[i].val_names[*menu[i].val - menu[i].low]);
			else
				printf("%13d", *menu[i].val);
		}
		else if (strlen(menu[i].name) > 10)
			printf("%*s", 13 - (strlen(menu[i].name) - 10), "");
		else
			printf("%13s", "");

		printf("%c", trs_menu ? 170 : 186);
		m++;
		y++;
	}

	gotoxy(x, y);
	botline(MENWID, trs_menu);
	y++;
}

void drawline(int, char, char, char);

static void topline(int width, int trs_menu)
{
	if (trs_menu)
		drawline(width, 156, 140, 172);
	else
		drawline(width, 201, 205, 187);
}

static void botline(int width, int trs_menu)
{
	if (trs_menu)
		drawline(width, 131, 131, 131);
	else
		drawline(width, 200, 205, 188);
}

static void drawline(int width, char c1, char c2, char c3)
{
	printf("%c", c1);

	for (width -= 2; width > 0; width--)
		printf("%c", c2);

	printf("%c", c3);
}

/* Change the value of a menu item and then change the emulator
 * state to reflect the new configuration.
 * Returns:
 *	0 if the menu should continue
 *	1 if the emulator should exit
 *	2 if the menu should disappear (and the emulator continue)
 */
static int menu_change(int sel, int add)
{
	int i, j, off;

	*menu[sel].val += add;

	if (*menu[sel].val > menu[sel].high)
		*menu[sel].val = menu[sel].low;

	if (*menu[sel].val < menu[sel].low)
		*menu[sel].val = menu[sel].high;

	/* handle special menu items */
	if (sel == M_QUIT)
		return 1;
	if (sel == M_RETURN)
		return 2;

	if (menu[sel].change)
		(*menu[sel].change)();

	return 0;
}
