/*
 * rconfig.c -- read the configuration file (trs80rc)
 *
 * The configuration file is pretty simple, comments are delimited with
 * #\n, the rest of the file is broken up into a stream of tokens
 * separated by white space.
 *
 * Keywords indicate if the current section of the file is for keyboard,
 * display, etc.
 */

#include <stdio.h>
#include <alloc.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"

#include "rconfig.h"
#include "load.h"

#define PATH	(2)

void read_config(fp)
FILE* fp;
{
	char*	t;
	int		section = PATH;

	while ((t = fgettok(fp)) != NULL) {
		if (!strcmp(t, "path"))
			section = PATH;
		else switch (section) {
			case PATH:	path_token(t); break;
		}
	}
	fclose(fp);
}

char* curline = NULL;
static char buf[256];

char* fgettok(FILE* fp)
{
	char*	p;
	char*	token;

	do {
		if (curline == NULL)
			if (fgets(buf, 256, fp) == NULL)
				return(NULL);
			else
				curline = buf;

		while (*curline == '\t' || *curline == ' ')
			curline++;
		if (*curline == '#' || *curline == '\n' || *curline == '\0')
			curline = NULL;
	} while (curline == NULL);

	token = curline;
	while (*curline != ' ' && *curline != '\t' && *curline != '\n' && *curline != '\0')
		curline++;
	*curline++ = '\0';

	return(token);
}

static char* clonebuf;
static int clonefree = 0;
#define CLONECHUNK	(1024)

/* clone a string. this tries to be fairly memory efficient; don't
 * free the space you get from this!
 */
char* clonestr(char* orig)
{
	int		len;
	char*	buf;

	if ((len = strlen(orig) + 1) > clonefree) {
		if ((clonebuf = malloc(CLONECHUNK)) == NULL)
			fatal("out of memory");
		if (len > CLONECHUNK)
			fatal("string too big");
		clonefree = CLONECHUNK;
	}

	strcpy(buf = clonebuf, orig);
	clonebuf += len;
	clonefree -= len;
	return(buf);
}
