/*
 * screen.c -- some of the device independent TRS-80 screen handling stuff
 *
 * This is mostly stuff for parsing configuration files and dealing with
 * virtual to real character set mappings.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <alloc.h>

#include "screen.h"
#include "rconfig.h"

struct charmap {
	char*	name;
	char*	map;
	struct charmap*	next;
};

static struct charmap* cmap_list = NULL;
static struct charmap** cmap_insert = &cmap_list;

int mapping[256];

char* char_stat();
void char_adjust(char*);
static int make_mapping(char*, int*);

static struct charmap* cur_map;

void screen_init()
{
	/* should make sure there are some mappings! */
	cur_map = cmap_list;
}

void screen_token(char* tok)
{
	char*	map;

	if ((map = strrchr(tok, ':')) == NULL)
		fatal("bad character mapping\n");

	*map++ = '\0';
	switch (make_mapping(map, mapping)) {
	case 0: break;
	case 1: fatal("character value out of range in mapping\n");
	case 2: fatal("no high end in character range\n");
	case 3: fatal("not enough characters in mapping\n");
	case 4: fatal("missing number in character set mapping\n");
	}

	if ((*cmap_insert = (struct charmap*)malloc(sizeof(struct charmap))) == NULL)
		fatal("out of memory\n");

	(*cmap_insert)->next = NULL;
	(*cmap_insert)->name = clonestr(tok);
	(*cmap_insert)->map = clonestr(map);

	cmap_insert = &((*cmap_insert)->next);
}

#define NUMCHAR (263)

int checkrange(int);

static int make_mapping(char* maparg, int* map)
{
	char*	p;
	int		i;
	int		cur = 0;
	char	ch;
	int		low;
	int		high;

	p = maparg;
	do {
		/* what the hey! */
		while (*p == ',')
			p++;

		if (!isdigit(*p))
			return(4);

		maparg = p;
		while (isdigit(*p))
			p++;

		ch = *p;
		*p = '\0';
		if (checkrange(high = low = atoi(maparg)))
			return(1);

		*p = ch;
		if (ch == '-') {
			p++;
			if (!isdigit(*p))
				return(2);

			maparg = p;
			while (isdigit(*p))
				p++;
			ch = *p;
			*p = '\0';
			if (checkrange(high = atoi(maparg)))
				return(1);
			*p = ch;
		}
		if (low < high) {
			for (i = low; i <= high; i++) {
				if (cur > 255)
					return(1);
				map[cur++] = i;
			}
		}
		else {
			for (i = low; i >= high; i--) {
				if (cur > 255)
					return(1);
				map[cur++] = i;
			}
		}
	} while (*p == ',');

	if (cur < 256)
		return(3);

	return(0);
}

int checkrange(int n)
{
	return (n < 0 || n >= NUMCHAR);
}
