/*
 * snap.h
 */

extern int savesnap(struct trs80*, char*);
extern int restoresnap(struct trs80*, char*);

/* Attempt at a generic emulator snapshot format.  The idea is to have
 * a well-known set of tags for data.  Not likely to be used by anyone
 * but me, but I think the idea is nice.  The format it two level.  A
 * magic number indicates it is an emulator snapshot.  It is immediately
 * followed by a number indicating the general hardware type.  Then
 * there is a series of tags which have pre-defined length (so they
 * may be skipped) and are broken into those specific to particular
 * hardware and universal ones.  Universal ones are used whenever
 * possible.
 *
 * Tags 0 -> 63 are universal with no extra data (i.e., length 0)
 * Tags 64 -> 127 are machine specific with length 0
 * Tags 128 -> 191 are universal with null character termination
 * Tags 192 -> 255 are machine specific with null character termination
 * Tags 256 -> 511 are universal with 1 byte length
 * Tags 512 -> 767 are machine specific with 1 byte length
 * Tags 768 -> 1023 are universal with 2 byte length
 * Tags 1024 -> 1279 are machine specific with 2 byte lengths
 * Tags 1280 -> 2047 are universal with 4 byte lengths
 * Tags 2048 -> 3071 are machine specific with 4 byte lengths
 *
 * All other tags have 4 byte lengths and are as yet undefined.
 */

/* Top level magic number identifier */
#define SNAP_MAGIC	"sn\301P"

/* hardware identifiers */
#define SNAP_MODEL_1_3	(1)
#define APPLE2			(2)
/* etc... */

/* universal no data tags */
#define SNAP_EOF	(0)
/* universal null terminated tags */
#define SNAP_CREATOR	(128)

/* universal one byte length tags */
/* written out in big endian in the following order:
	AF BC DE HL AF' BC' DE' HL' IX IY SP PC I R IM IFF1 IFF2 */
#define SNAP_Z80	(256)
#define SNAP_6502	(257)	/* to be defined */

/* universal two byte length tags */
/* 2 byte big-endian start address with remainder of tag data the
 * bytes to load into memory.
 */
#define SNAP_MEM16	(768)
