/*
 * trs80pc.c -- routines for TRS-80 hardware emulation on a PC
 *
 * Big todo: put z80 cpu state into a separate structure or something
 * to make life easier for the emulators.  The copy-in operation
 * is somewhat painful.
 */

#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <graphics.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "refz80.h"
#include "trs80pc.h"
#include "key.h"
#include "clock.h"

#include "t_to_pc.h"

extern int cga;
extern int text;

/* menu changeable configuration parameters */
int charset = 1;
int physmap = 0;
int sounds = 1;
int colour = 7; /* white */
int speed = 1;

int doublewide = 0;
int katana = 0;
mbyte* screen_mem;
static int screen_type;
static int do_screen_update = 0;
static int trs_menu;
int fullupdate = 0;
extern int simplemode;
long MHz = MODEL3_MHZ;

mbyte trs_char32_glyph[sizeof(trs_char_glyph)*2];

static int trs_charmap[256];
mbyte trs_char[256*12];
mbyte trs_char32[256*12*2];

#define VGA_TEXT	(0)
#define CGA_GRAPH	(1)
#define DUMB_TEXT	(2)

/* A kludge until we get the overall structure more organized */
/* which is almost there with the struct z80 stuff */
static struct z80* sz80;

static void port255(mbyte, mbyte);
static void port_ec(mbyte, mbyte);

void init_trs80pc(struct trs80* t)
{
	wportcall_z80(&t->z80, 255, port255);
	wportcall_z80(&t->z80, 0xec, port_ec);
	sz80 = &t->z80;
	init_scrn(t);
}

static void port255(mbyte port, mbyte val)
{
	mbyte p61;
	struct REGPACK r;

	if (sounds)
		outportb(0x61, (inportb(0x61) & 0xfc) | ((val & 1) << 1));

	doublewide = !!(val & 8);
}

static void port_ec(mbyte port, mbyte val)
{
	doublewide = !!(val & 4);

	/* According to the manual, this is not right.  However, it
	 * appears to be the way it should be */
	katana = !(val & 8);
}

static int curdw = 0;
static int curkatana = 0;

void screen_refresh()
{
	char far* scrn = MK_FP(0xb800, 0);
	char far* ts;
	int row, col;

	if (!do_screen_update)
		return;

	if (curdw != doublewide || curkatana != katana) {
		curdw = doublewide;
		curkatana = katana;

		if (charset == 1) {
			int off, i;

			off = katana ? 71 : 0;
			for (i = 192; i < 256; i++)
				trs_charmap[i] = i + off;
		}

		setmode();
	}

	switch (screen_type) {
	case CGA_GRAPH:
		if (doublewide)
			cgascreen32();
		else
			cgascreen64();
		break;
	case VGA_TEXT:
		ts = screen_mem;
		if (doublewide) {
			scrn += 4;
			for (row = 0; row < 16; row++) {
				for (col = 0; col < 32; col++) {
					*scrn++ = *ts++;
					*scrn++ = colour;
					ts++;
				}
				scrn += (40 - 32) * 2;
			}
		}
		else {
			scrn += 8;
			for (row = 0; row < 16; row++) {
				for (col = 0; col < 64; col++) {
					*scrn++ = *ts++;
					*scrn++ = colour;
				}
				scrn += (80 - 64) * 2;
			}
		}
		break;
	case DUMB_TEXT:
		/* I could do better, but this mode is only for debugging */
		ts = screen_mem;
		if (doublewide) {
			for (row = 0; row < 16; row++) {
				for (col = 0; col < 32; col++) {
					if (*ts & 128)
						*scrn++ = t_to_pc[*ts++ & 63];
					else
						*scrn++ = *ts++;
					*scrn++ = colour;
					ts++;
				}
				scrn += (40 - 32) * 2;
			}
		}
		else {
			for (row = 0; row < 16; row++) {
				for (col = 0; col < 64; col++) {
					if (*ts & 128)
						*scrn++ = t_to_pc[*ts++ & 63];
					else
						*scrn++ = *ts++;
					*scrn++ = colour;
				}
				scrn += (80 - 64) * 2;
			}
		}
		break;
	default:
		break;
	}
}

static void dump_reg(struct z80*, int);
static int htoi(char*);

void run_trs80(struct trs80* t)
{
	static char dis[64];
	mword npc, i, opc, daddr = 0;
	int singlestep, key, blank, vloop, r, vr, loadup;
	long cycperslice, cycused, cycsave;

	do_screen_update = 1;

	singlestep = 0;
	/*singlestep = 1;*/

	if (singlestep) goto debug;

	loadup = 1;
	for (;;) {
		/* Figure out how many cycles per 1/19 second time slice at x MHz */
		cycperslice = MHz / 19;
		cycused = 0;

		kbinstall(t);
		if (t->ref) {
			t->ref->z80.reg = t->z80.reg;
		}
		for (vloop = 1; vloop; ) {
			if (t->ref) {
				opc = t->z80.reg.PC;
				/*dump_reg(&t->z80, 0);*/
			}
			cycsave = t->z80.cycles;
			t->z80.cycles = 0;
			r = run_z80(&t->z80, loadup, singlestep || t->ref);
			cycused += t->z80.cycles;
			t->z80.cycles += cycsave;
			if (MHz && cycused >= cycperslice) {
				while (!did_intr)
					;
				did_intr = 0;
				cycused -= cycperslice;
			}
			if (esc_hit) {
				vloop = 0;
				esc_hit = 0; /* extremely small window where we could miss an ESC */
			}
			switch (r) {
			case PROG_STOP:
				vloop = 0;
				break;
			case ILL_INST:
				sprintf(merr, "Illegal instruction at %04x", t->z80.reg.PC);
				vloop = 0;
				break;
			case HALT:
				sprintf(merr, "Halt at %04x", t->z80.reg.PC);
				vloop = 0;
				break;
			case TIME_OVER:
				/* speed control! */
				if (singlestep)
					vloop = 0;
				break;
			default:
				sprintf(merr, "Internal error");
				vloop = 0;
				break;
			}
			if (t->ref) {
				vr = refz80(&t->ref->z80);
				/* XXX for now, mask the flag register until I determine
				 * what's the right emulation of the "don't care" bits.
				 */
				t->z80.reg.AF &= 0xffd7;
				t->ref->z80.reg.AF &= 0xffd7;
				if (memcmp(&t->z80.reg, &t->ref->z80.reg, sizeof(struct regz80))) {
					printf("deviation from reference model at %04x.\n", opc);
					vloop = 0;
					dump_reg(&t->ref->z80, 1);
					dump_reg(&t->z80, 0);
					t->z80.reg.PC = opc;
				}
			}
			/* Does this have a function? */
			/*else
				vloop = 0;*/

			loadup = 0;
		}
		/* Turn off sound; if the timing is exactly right, the speaker
		 * could be in beep mode right now and that would be very annoying.
		 */
		outportb(0x61, inportb(0x61) & 0xfc);
		kbremove();
		if (simplemode) {
			do_screen_update = 0;

			if (do_menu(screen_type == CGA_GRAPH ? 0 : doublewide, trs_menu))
				return;

			/* very cheap way to erase the menu */
			setmode();
			do_screen_update = 1;

			continue;
		}

	debug:
		npc = disasm(&t->z80, t->z80.reg.PC, dis);
		printf("%04x: ", t->z80.reg.PC);
		blank = 4;
		for (i = t->z80.reg.PC; i < npc; i++) {
			printf("%02x ", t->z80.mem[i]);
			blank--;
		}
		while (blank-- > 0)
			printf("   ");
		printf("%s\n", dis);

		printf(">\r");
		for (key = 0; !key; ) {
			switch (getch()) {
			case 'q': case 'Q':
				return;
			case 's': case 'S':
				printf("step...\r");
				singlestep = 1;
				key = 1;
				break;
			case 'c': case 'C':
				printf("continuing...\r");
				singlestep = 0;
				key = 1;
				break;
			case 'r': case 'R':
				dump_reg(&t->z80, 1);
				printf(">\r");
				break;
			case 'd': case 'D':
				printf(">d ");
				fgets(dis, 64, stdin);
				daddr = htoi(dis);
				printf("\r%04x:", daddr);
				for (i = 0; i < 8; i++) {
					printf(" %02x%02x", t->z80.mem[daddr], t->z80.mem[daddr + 1]);
					daddr += 2;
				}
				printf("\n>\r");
				break;
			case 'p': case 'P':
				printf(">snapshot file? ");
				fgets(dis, 64, stdin);
				if (strchr(dis, '\n')) *strchr(dis, '\n') = '\0';
				savesnap(dis);
				printf(">\r");
			default:
				printf("(Q)uit, (S)inglestep, (C)ontinue, (D)ump, (R)egisters, sna(P)shot\n");
				break;
			}
		}
	}
}


static void dump_reg(struct z80* z, int names)
{
	if (names)
		printf(" PC   AF   BC   DE   HL   SP   IX   IY   AF'  BC'  DE'  HL' I  R  IM IF1 IF2\n");

	printf("%04x %04x %04x %04x %04x %04x %04x %04x %04x %04x %04x %04x %02x %02x %x  %x   %x\n",
		z->reg.PC, z->reg.AF, z->reg.BC, z->reg.DE, z->reg.HL, z->reg.SP,
		z->reg.IX, z->reg.IY, z->reg.AFp, z->reg.BCp, z->reg.DEp, z->reg.HLp,
		z->reg.I, z->reg.R, z->reg.IM, z->reg.IFF1, z->reg.IFF2);
}

static int htoi(char* str)
{
	int val;

	while (*str == ' ' && *str == '\t')
		str++;

	val = 0;

	while ((*str >= '0' && *str <= '9') ||
		   (*str >= 'a' && *str <= 'f') ||
		   (*str >= 'A' && *str <= 'F'))
	{
		val <<= 4;
		if (*str < 'A')
			val |= *str - '0';
		else if (*str < 'a')
			val |= *str - 'A' + 10;
		else
			val |= *str - 'a' + 10;

		str++;
	}

	return val;
}

static int inited = 0;

init_scrn(struct trs80* t)
{
	int gdriver, gmode, err, i;

	do_screen_update = 0;

	for (i = 0; i < 256; i++)
		trs_charmap[i] = i;

	screen_mem = t->z80.mem + 15360;

	if (registerbgidriver(CGA_driver) < 0) {
		fprintf(stderr, "can't register cga driver!?\n");
		exit(1);
	}

	if (registerbgidriver(Herc_driver) < 0) {
		fprintf(stderr, "can't register herc driver!?\n");
		exit(1);
	}

	gdriver = DETECT; /* is this necessary? */
	detectgraph(&gdriver, &gmode);
	err = graphresult();
	if (err != grOk) {
		fprintf(stderr, "detectgraph: %s\n", grapherrormsg(err));
		exit(1);
	}

	switch (gdriver) {
	case MCGA:
	case CGA:
	case EGA:
	case EGAMONO:
	case EGA64:
		gdriver = CGA;
		gmode = CGAHI;
		break;
	case HERCMONO:
		gdriver = HERCMONO;
		break;
	case VGA:	/* maybe EGA can handle what I want, too, I dunno */
		break;
	default:
		fprintf(stderr, "Sorry, I need CGA, Hercules mono or VGA.\n");
		exit(1);
	}

	if (cga && gdriver == VGA) {
		gdriver = CGA;
		gmode = CGAHI;
	}

	if (text) {
		screen_type = DUMB_TEXT;
		setmode();
	}
	else if (gdriver == VGA) {
		screen_type = VGA_TEXT;
		setmode();
	}
	else if (gdriver != CGA) {
		printf("Sorry, gotta have CGA or VGA.\n");
		exit(1);
	}
	else {
		mbyte *src, *dst, dbl[16];
		int i;

		initgraph(&gdriver, &gmode, ".");
		err = graphresult();

		if (err != grOk) {
			fprintf(stderr, "Graphics error: %s\n", grapherrormsg(err));
			exit(1);
		}
		screen_type = CGA_GRAPH;

		dbl[ 0] = 0x00; dbl[ 1] = 0x03; dbl[ 2] = 0x0c; dbl[ 3] = 0x0f;
		dbl[ 4] = 0x30; dbl[ 5] = 0x33; dbl[ 6] = 0x3c; dbl[ 7] = 0x3f;
		dbl[ 8] = 0xc0; dbl[ 9] = 0xc3; dbl[10] = 0xcc; dbl[11] = 0xcf;
		dbl[12] = 0xf0; dbl[13] = 0xf3; dbl[14] = 0xfc; dbl[15] = 0xff;

		for (src = trs_char_glyph, dst = trs_char32_glyph;
			 src < trs_char_glyph + sizeof(trs_char_glyph);
			 src++)
		{
			*dst++ = dbl[*src >> 4];
			*dst++ = dbl[*src & 15];
		}
		menu_active(M_COLOUR, 0);
	}
	build_charmap();

	inited = 1;
}

build_charmap()
{
	int i;

	for (i = 0; i < 256; i++)
		memcpy(trs_char + i * 12, trs_char_glyph + trs_charmap[i] * 12, 12);

	if (screen_type == CGA_GRAPH) {
		for (i = 0; i < 256; i++)
			memcpy(trs_char32 + i * 24,
				   trs_char32_glyph + trs_charmap[i] * 24, 24);
	}
}


setmode()
{
	struct REGPACK r;

	build_charmap();

	switch (screen_type) {
	case VGA_TEXT:
		r.r_ax = 0x1200;
		r.r_bx = 0x0030;
		intr(0x10, &r);		/* select 200 line mode */

		r.r_ax = doublewide ? 0x0001 : 0x0003;
		intr(0x10, &r);

		instcharrom();

		cursoroff();

		trs_menu = 1;
		break;
	case DUMB_TEXT:
		r.r_ax = doublewide ? 0x0001 : 0x0003;
		intr(0x10, &r);
		cursoroff();
		trs_menu = 0;
		break;
	default:
		trs_menu = 0;
		break;
	}


	fullupdate = 1;
}

cursoroff()
{
	struct REGPACK r;

	r.r_ax = 0x0100;
	r.r_cx = 0x2000;
	intr(0x10, &r);		/* turn off the cursor */
						/* a printf() will turn in on so be careful */
}

void fixpc(void)
{
	struct REGPACK r;

	if (inited) {
		switch (screen_type) {
		case VGA_TEXT:
			r.r_ax = 0x1202;
			r.r_bx = 0x0030;
			intr(0x10, &r);				/* select 400 line mode */
			/* fall through */
		case DUMB_TEXT:	/* to get cursor back on */
			r.r_ax = 0x0003;
			intr(0x10, &r);				/* set standard 80x25 mode */
			break;
		case CGA_GRAPH:
			closegraph();
			break;
		default:
			break;
		}
	}
	inited = 0;
}

mbyte shadow[64*16];

void change_colour(void)
{
	char far* s = MK_FP(0xb800, 1);
	int len = doublewide ? 16 * 40 : 16 * 80;

	while (len--) {
		*s++ = colour;
		s++;
	}
}

void change_charset(void)
{
	int i, j, off;

#define mapin(x)	trs_charmap[i++] = x
#define maprange(x,y)	for (j = x; j <= y; j++) mapin(j);

	switch (charset) {
	case 0:	/* Standard Model I character set */
		i = 0;
		maprange(64, 90);
		maprange(256, 259)
		mapin(95);
		maprange(32, 90);
		maprange(256, 259);
		mapin(95);
		mapin(260);
		maprange(65, 90);
		maprange(123, 125);
		maprange(261, 262);
		maprange(128, 191);
		maprange(128, 191);
		break;
	case 1:	/* Standard Model III character set */
		for (i = 0; i < 192; i++)
			trs_charmap[i] = i;

		off = katana ? 71 : 0;
		for (i = 192; i < 256; i++)
			trs_charmap[i] = i + off;

		break;
	case 2:	/* Model I with lower case enhancement character set */
		i = 0;
		maprange(64, 90);
		maprange(256, 259);
		mapin(95);
		maprange(32, 90);
		maprange(256, 259);
		mapin(95);
		mapin(260);
		maprange(97, 125);
		maprange(261, 262);
		maprange(128, 191);
		maprange(128, 191);
		break;
	}
	setmode();
	i = do_screen_update;
	do_screen_update = 1;
	screen_refresh();
	do_screen_update = i;
}

void change_speed(void)
{
	switch (speed) {
	case 0: MHz = MODEL1_MHZ; break;
	case 1: MHz = MODEL3_MHZ; break;
	case 2: MHz = 0; break;
	case 3: MHz = MODEL1_MHZ / 2; break;
	}
}
