/*
 * trs83.c -- a trs80 simulator!
 *
 * Hmmm, routines should not call fatal if the system has not yet
 * been installed.  Looks like "key.c" and "load.c" have to be changed!
 *
 * Author: George Phillips <phillips@cs.ubc.ca>
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include "mtype.h"
#include "z80.h"
#include "trs80.h"
#include "trs80pc.h"
#include "load.h"
#include "key.h"
#include "clock.h"
#include "config.h"
#include "rconfig.h"
#include "screen.h"
#include "snap.h"

int cga = 0;
int text = 0;
/*extern*/ long icount;
int simplemode = 1;

void check_z80();

struct trs80 trs80;

void main(int argc, char* argv[])
{
	int	i;
	int ret;
	int gotfile;
	static struct trs80 vtrs80;
	char* errmess;
	int diskwaitkludge;
	FILE* fp;
	static char buf[256];
	static char* arg[64];

	static char mess[128];

	printf("TRS-80 Simulator by George + Peter Phillips\n");
	if (argc == 1) {
		printf("usage: trs83 [-1 -3 -cga] trs80 programs\n");
		exit(0);
	}

	check_z80();

	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		if (!strcmp(argv[i], "-v")) {
			init_z80(&vtrs80.z80);
			vtrs80.ref = 0;
			trs80.ref = &vtrs80;
		}
		else if (!strcmp(argv[i], "-1")) { speed = 0; }
		else if (!strcmp(argv[i], "-3")) { speed = 1; }
		else if (!strcmp(argv[i], "-g")) { simplemode = 0; }
		else if (!strcmp(argv[i], "-cga")) { cga = 1; }
		else if (!strcmp(argv[i], "-text")) { text = 1; }
		else {
			printf("unknown option %s\n", argv[i]);
			exit(1);
		}
	}

	if (!init_z80(&trs80.z80)) {
		printf("initialization failed (out of memory?)\n");
		exit(1);
	}
	trs80.ref = 0;
	init_trs80pc(&trs80);
	initrefz80();

	diskwaitkludge = 0;

	if ((fp = fopen("trs80rc", "r")) == NULL) {
		char* p;

		strcpy(buf, argv[0]);
		if (p = strrchr(buf, '\\'))
			p++;
		else
			p = buf;

		strcpy(p, "trs80rc");
		fp = fopen(buf, "r");
	}

	if (fp)
		read_config(fp);

	gotfile = 0;
	for (; i < argc; i++) {
		gotfile = 1;
		/*printf("loading %s\n", argv[i]);*/
		if (!load(&trs80, argv[i])) {
			shutdown();
			printf("Error loading file '%s'\n", argv[i]);
			if (loaderr_str[0])
				printf("%s\n", loaderr_str);
			if (loaderr & LD_SYSERR)
				perror("System error was");
			exit(1);
		}
	}

	screen_init();

	icount = 0;
	kbinstall(&trs80);

	/* turn off the interrupts initially, prob. should be in z80init() */
	/*int_enabled = 0;*/

	if (diskwaitkludge)
		sleep(5);

	if (trs80.ref) {
		memcpy(vtrs80.z80.mem, trs80.z80.mem, 32768);
		memcpy(vtrs80.z80.mem + 32768, trs80.z80.mem + 32768, 32768);
	}

	clock_install();

	run_trs80(&trs80);
	fatal("Bye now!\n");
}

extern long numi[256];
extern int key_was_running;
extern int key_interrupts;
extern int key_repeated;
extern int key_during_screen;
extern int refresh_overrun;

shutdown()
{
	kbremove();
	clock_remove();
	fixpc();

	/* turn off sound! */
	outportb(0x61, inportb(0x61) & 0xfc);
}

void fatal(char* s)
{
	FILE*	fp;
	int		i;
	long	num_inst;
	float	et = (float)clock_tick / clock_freq;
	/* of course, the clock_tick above is quite useless if the SUPS is changed! */

	shutdown();

	fprintf(stderr, "%ld cycles in %f seconds\n",
		trs80.z80.cycles, (float)clock_tick / clock_freq);

	if (trs80.z80.cycles > 0 && et > 0)
		fprintf(stderr, "%f cycles per second, i.e., %f MHz\n",
			(float)trs80.z80.cycles / et,
			(float)trs80.z80.cycles / et / 1000000.0);

	fprintf(stderr, "%s\n", s);

#ifdef hacker
	fprintf(stderr, "PC: %-4.4x\n", trs80.z80.reg.PC);
	fprintf(stderr, "AF: %-4.4x\n", trs80.z80.reg.AF);
	fprintf(stderr, "BC: %-4.4x\n", trs80.z80.reg.BC);
	fprintf(stderr, "DE: %-4.4x\n", trs80.z80.reg.DE);
	fprintf(stderr, "HL: %-4.4x\n", trs80.z80.reg.HL);

	fprintf(stderr, "%d keyboard interrupts occured:\n", key_interrupts);
	fprintf(stderr, "%d overlapping\n", key_was_running);
	fprintf(stderr, "%d repeats\n", key_repeated);
	fprintf(stderr, "%d during screen refresh\n", key_during_screen);
	fprintf(stderr, "%d screen update overruns\n", refresh_overrun);
#endif

	exit(0);
}

mkarg(s, arg)
char* s;
char* arg[];
{
	int n = 0;

	while (*s) {
		while (isspace(*s))
			*s++ = '\0';

		if (*s)
			arg[n++] = s;

		while (*s && !isspace(*s))
			s++;
	}
	arg[n] = 0;
	return n;
}

/*
 * A "production" version does not need this routine.  However, it
 * provides an important sanity check.  Because of the funky decoding
 * used by the z80 simulator, there is some magic to the code placement.
 * This checks that the bits of code which implement each instruction
 * don't overlap.
 */

extern int istart[], iend[];

void check_z80()
{
	int i, j, off[8], len[8], l, prev;

	for (j = 0; j < 5; j++) {
		off[j] = istart[j * 256] & 255;
		len[j] = 0;
		for (i = 0; i < 256; i++) {
			if ((istart[j * 256 + i] & 255) != off[j])
				fprintf(stderr, "series %d has a misaligned opcode (%d)\n", j, i);
			l = iend[j * 256 + i] - istart[j * 256 + i];
			if (l > len[j])
				len[j] = l;
		}
	}

#ifdef hacker
	for (j = 0; j < 5; j++) {
		printf("%3d %3d %3d\n", off[j], len[j], off[j] + len[j]);
	}
#endif

	prev = off[0] + len[0];
	for (j = 1; j < 5; j++) {
		if (off[j] < prev) {
			fprintf(stderr, "series %d collision\n", j);
			exit(1);
		}
		prev = off[j] + len[j];
	}
}
