/*
 * z80.h -- abstract interface to a z80 simulator
 */


/* This structure should be very standard across z80 simulators */
struct regz80 {
	/* You can play with these register values */
	mword AF, BC, DE, HL, IX, IY, SP, PC, AFp, BCp, DEp, HLp;
	/* Best to consider these read only */
	mbyte I, R, IM, IFF1, IFF2;
};

/* The "struct z80" is an abstract type, except for the first
 * couple fields.
 */

typedef mbyte (*memcall)(mword, mbyte, mbyte*);
typedef mbyte (*portread)(mbyte);
typedef void (*portwrite)(mbyte, mbyte);

struct calldemux {
	memcall* mux;
};

struct z80 {
	struct regz80 reg;
	long cycles;
	mbyte far* mem;
	/* Don't touch this stuff, for internal use only */
	portread port_read[256];
	portwrite port_write[256];
	memcall *mem_write[256];
	memcall *mem_read[256];
};

int init_z80(struct z80*);
int run_z80(struct z80*, int, int);
void stop_z80();
void rportcall_z80(struct z80*, mbyte, portread);
void wportcall_z80(struct z80*, mbyte, portwrite);

/* return codes from run_z80 */
#define PROG_STOP	(0)	/* Z80 halted 'cause program called stop_z80() */
#define ILL_INST	(1)	/* illegal instruction encountered */
#define	HALT		(2)	/* HALT instruction encountered */
#define TIME_OVER	(3)	/* Z80 ran out of cycles allocated to it */


